/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import CoreBluetooth

class OADDevice
{
    var name = ""
    var startAddr: UInt32 = 0x0
    var maxImageSize: UInt32 = 262144
    
    init(devName: String, startAddress: UInt32, maxImage: UInt32)
    {
        self.name = devName
        self.startAddr = startAddress
        self.maxImageSize = maxImage
    }
}

let oadDevices =
    [
        OADDevice(devName: "MSP432P401x", startAddress: 0x00020000, maxImage: 262144),
        OADDevice(devName: "MSP432P4111", startAddress: 0x00100000, maxImage: 1048576),
        OADDevice(devName: "Custom", startAddress: 0x00020000, maxImage: 0xFFFFFFFF),
]


class MSP432FirmwareDownloadViewController: UIViewController, UITextFieldDelegate, UIPickerViewDelegate,
    UITableViewDataSource, UITableViewDelegate, UIDocumentPickerDelegate, CBPeripheralDelegate,
    CBCentralManagerDelegate, UIPickerViewDataSource
{
    
    // MARK: Properties
    @IBOutlet weak var firmwareTitle: UILabel!
    @IBOutlet weak var browseButton: UIButton!
    @IBOutlet weak var deviceLabel: UILabel!
    @IBOutlet weak var devicePicker: UIPickerView!
    @IBOutlet weak var addressLabel: UILabel!
    @IBOutlet weak var titleLabel: UILabel!
    @IBOutlet weak var addressField: UITextField!
    @IBOutlet weak var firmwareNameLabel: UILabel!
    @IBOutlet weak var disconButton: UIButton!
    @IBOutlet weak var firmwareView: UIView!
    @IBOutlet weak var optionsLabel: UILabel!
    @IBOutlet weak var versionLabel: UILabel!
    @IBOutlet weak var bleView: UIView!
    @IBOutlet weak var versionField: UITextField!
    @IBOutlet weak var optionsView: UIView!
    @IBOutlet weak var deviceTitle: UILabel!
    @IBOutlet weak var sapLabel: UILabel!
    @IBOutlet weak var deviceTable: UITableView!
    @IBOutlet weak var programButton: UIButton!
    @IBOutlet weak var imageTypePicker: UISwitch!
    @IBOutlet weak var snpLabel: UILabel!
    
    // MARK: UUIDs for OAD Service
    let TI_OAD_SERVICE_UUID = CBUUID(string: "F000FFC0-0451-4000-B000-000000000000")
    let TI_OAD_IMG_IDENTIFY_UUID = CBUUID(string: "F000FFC1-0451-4000-B000-000000000000")
    let TI_OAD_IMG_BLOCK_UUID = CBUUID(string: "F000FFC2-0451-4000-B000-000000000000")
    let TI_OAD_EXT_CTL_UUID = CBUUID(string: "F000FFC5-0451-4000-B000-000000000000")
    
    // MARK: Bluetooth Characteristics
    var imgIdentChar: CBCharacteristic!
    var imgBlockChar: CBCharacteristic!
    var imgExtCtlChar: CBCharacteristic!
    
    // MARK: Variables
    var curFirmware: ImportedFirmwareImage!
    var firmwareManager = OADFirmwareManager()
    var firmwareDownloadView: StatusLogViewController!
    var waitWriteSem: DispatchSemaphore!
    var blockWriteSem: DispatchSemaphore!
    var waitForCommandSem: DispatchSemaphore!
    var metaDataRejected: Bool = false
    var writeCommandFailed: Bool = false
    var curBlockSize: Int = FIRMWARE_OAD_MAX_PAYLOAD_SIZE
    var curPacketSize: Int = FIRMWARE_OAD_MAX_PAYLOAD_SIZE
    var connectionCon: DiscoveryViewController!
    var timeoutTimer: Timer!
    var cancelInvoked: Bool = false
    var btManager: CBCentralManager!
    var selBTDevs: [CBPeripheral]! = []
    var selBTDev: CBPeripheral!
    var rssiReadings: [NSNumber]! = []
    var advData: [NSData]! = []
    let hexSet = NSCharacterSet(charactersIn: "0123456789ABCDEFabcdef").inverted
    let identifier = "BLEDeviceTableViewCell"
    
    // MARK: Actions
    @IBAction func browsePressed(_ sender: UIButton)
    {
        /* Creating a document picker and letting the user choose their firmware image */
        let docs = UIDocumentPickerViewController(documentTypes: ["public.item"],
                                                  in: UIDocumentPickerMode.import)
        docs.delegate = self
        docs.modalPresentationStyle = UIModalPresentationStyle.fullScreen
        self.present(docs, animated: true, completion: nil)
    }
    
    @IBAction func privacyClicked(_ sender: Any)
    {
        UIApplication.shared.open(URL(string: "http://www.ti.com/corp/docs/legal/privacy.shtml")!,
                                  options: [:], completionHandler: nil)
    }
    
    
    @IBAction func termsClicked(_ sender: Any)
    {
        UIApplication.shared.open(URL(string: "http://www.ti.com/corp/docs/legal/termsofuse.shtml")!,
                                  options: [:], completionHandler: nil)
    }
    
    @IBAction func programPressed(_ sender: UIButton)
    {
        if(curFirmware == nil)
        {
            let alert = UIAlertController(title: "No Firmware Selected",
                                          message: "Please select firmware before initiating update.",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        if(selBTDev == nil)
        {
            let alert = UIAlertController(title: "Not Connected",
                                          message: "Please connect to a BLE device before initiating update.",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        if((addressField.text?.isEmpty)! || (versionField.text?.isEmpty)!)
        {
            let alert = UIAlertController(title: "Invalid Values",
                                          message: "Invalid values for version/address",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        versionField.resignFirstResponder()
        addressField.resignFirstResponder()
        UIApplication.shared.isIdleTimerDisabled = true
        
        self.cancelInvoked = false
        self.performSegue(withIdentifier: "startDownloadSegue", sender: self)
    }
    
    @IBAction func typeSwitchPressed(_ sender: Any)
    {
        if(imageTypePicker.isOn)
        {
            sapLabel.font = UIFont.boldSystemFont(ofSize: 16.0)
            snpLabel.font  = UIFont.systemFont(ofSize: 16.0)
        }
        else
        {
            snpLabel.font = UIFont.boldSystemFont(ofSize: 16.0)
            sapLabel.font  = UIFont.systemFont(ofSize: 16.0)
        }
    }
    
    @IBAction func disconnectPressed(_ sender: Any)
    {
        btManager.cancelPeripheralConnection(selBTDev)
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        /* Adjusting various GUI elements */
        programButton.clipsToBounds = true
        programButton.layer.cornerRadius = 8.0
        browseButton.clipsToBounds = true
        browseButton.layer.cornerRadius = 8.0
        disconButton.clipsToBounds = true
        disconButton.layer.cornerRadius = 8.0
        snpLabel.adjustsFontSizeToFitWidth = true
        sapLabel.adjustsFontSizeToFitWidth = true
        firmwareTitle.clipsToBounds = true
        firmwareTitle.layer.cornerRadius = 8.0
        titleLabel.adjustsFontSizeToFitWidth = true
        firmwareView.clipsToBounds = true
        firmwareView.layer.cornerRadius = 8.0
        optionsView.clipsToBounds = true
        optionsView.layer.cornerRadius = 8.0
        optionsLabel.clipsToBounds = true
        optionsLabel.layer.cornerRadius = 8.0
        bleView.clipsToBounds = true
        bleView.layer.cornerRadius = 8.0
        deviceTitle.clipsToBounds = true
        deviceTitle.layer.cornerRadius = 8.0
        firmwareNameLabel.adjustsFontSizeToFitWidth = true

        /* Setting delegates/data sources */
        deviceTable.dataSource = self
        deviceTable.delegate = self
        devicePicker.delegate = self
        devicePicker.dataSource = self
        versionField.delegate = self
        addressField.delegate = self
        
        /* Initializing the BT Manager */
        btManager = CBCentralManager(delegate: self, queue: nil)
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // MARK: CBCentralManagerDelegate
    func centralManagerDidUpdateState(_ central: CBCentralManager)
    {
        if central.state == .poweredOn
        {
            central.scanForPeripherals(withServices: nil, options: nil)
        }
        else
        {
            displayNoBluetooth()
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDiscover peripheral: CBPeripheral,
                        advertisementData: [String : Any],
                        rssi RSSI: NSNumber)
    {
        if(!selBTDevs.contains(peripheral))
        {
            var insertRow = 0
            
            while(insertRow < rssiReadings.count)
            {
                if(rssiReadings[insertRow].doubleValue < RSSI.doubleValue)
                {
                    break
                }
                
                insertRow = insertRow + 1
            }
            
            let newIndexPath = IndexPath(row: insertRow, section: 0)
            selBTDevs.insert(peripheral, at: insertRow)
            rssiReadings.insert(RSSI, at: insertRow)
            
            if let manData = advertisementData[CBAdvertisementDataManufacturerDataKey] as? NSData
            {
                advData.insert(manData, at: insertRow)
            }
            else
            {
                advData.insert(NSData(), at: insertRow)
            }
            
            deviceTable.insertRows(at: [newIndexPath], with: .bottom)
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didConnect peripheral: CBPeripheral)
    {
        let maxMTU: Int = selBTDev.maximumWriteValueLength(for: .withResponse)
        print("MTU: " + String(maxMTU))
        selBTDev.delegate = self
        selBTDev.discoverServices(nil)
        btManager.stopScan()
        deviceTitle.text = "Connected: \(selBTDev.name!)"
        deviceTable.isUserInteractionEnabled = false
        disconButton.isHidden = false
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDisconnectPeripheral peripheral: CBPeripheral,
                        error: Error?)
    {
        displayDisconnect()
        deviceTable.isUserInteractionEnabled = true
        disconButton.isHidden = true
        deviceTitle.text = "Device"
        selBTDev = nil
        btManager.scanForPeripherals(withServices: nil, options: nil)
        rssiReadings.removeAll()
        selBTDevs.removeAll()
        deviceTable.reloadData()
    }
    
    func displayDisconnect()
    {
        let alert = UIAlertController(title: "Disconnected",
                                      message: "Device was disconnected",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: self.completionDisconnect))
        self.present(alert, animated: true, completion: nil)
    }
    
    func completionDisconnect(action: UIAlertAction)
    {
        _ = self.navigationController?.popToViewController(self, animated: false)
        _ = self.navigationController?.popViewController(animated: true)
    }
    
    func displayNoBluetooth()
    {
        let alert = UIAlertController(title: "Bluetooth Disabled",
                                      message: "Bluetooth is disabled. Please enable Bluetooth and try again.",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: self.completionDisconnect))
        self.present(alert, animated: true, completion: nil)
    }
    
    // MARK: TableViewDataSource
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int
    {
        return selBTDevs.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell
    {
        let cell: BLEDeviceTableViewCell = tableView.dequeueReusableCell(withIdentifier: identifier, for:
            indexPath) as! BLEDeviceTableViewCell
        let device = selBTDevs[indexPath.row]
        let devName = device.name
        
        if devName != nil
        {
            cell.deviceText.text = device.name
        }
        else
        {
            cell.deviceText.text = "Unnamed Device"
        }
        
        cell.setSignalStrengthImage(rssi: rssiReadings[indexPath.row].floatValue)
        
        return cell
    }
    
    // MARK: Table view data source
    func numberOfSections(in tableView: UITableView) -> Int
    {
        return 1
    }
    
    func tableView(_ tableView: UITableView,
                            didSelectRowAt indexPath: IndexPath)
    {
        connectionCon = DiscoveryViewController(message: "Connecting...")
        present(connectionCon, animated: true, completion: nil)
        
        timeoutTimer = Timer.scheduledTimer(timeInterval: 5.0,
                                            target: self,
                                            selector: #selector(self.completionTimeout),
                                            userInfo: nil,
                                            repeats: false)
        
        selBTDev = selBTDevs[indexPath.row]
        btManager.connect(selBTDev, options: nil)
    }
    
    @objc func completionTimeout()
    {
        btManager.cancelPeripheralConnection(selBTDev)
        connectionCon.dismiss(animated: true, completion: nil)
        
        let alert = UIAlertController(title: "Connection Timeout",
                                      message: "Timeout occurred while connecting to device. Please verify correct" +
            " services are being advertised on device.",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: nil))
        _ = navigationController?.popToViewController(self, animated: true)
        present(alert, animated: true, completion: nil)
    }
    
    // MARK: UIDocumentPickerDelegate
    func documentPicker(_ controller: UIDocumentPickerViewController, didPickDocumentsAt urls: [URL])
    {
        do
        {
            /* Only the first firmware image is chosen. */
            let dataContents: Data = try Data(contentsOf: urls[0], options: NSData.ReadingOptions.uncached)
            let importedImage: ImportedFirmwareImage = ImportedFirmwareImage(firmware: dataContents,
                                                                             filename: urls[0].lastPathComponent)
            curFirmware = importedImage
            firmwareNameLabel.text = importedImage.getTableLabel()
            
        }
        catch
        {
            print("Data Read error.")
        }
    }
    
    // MARK: UIPickerViewDelegate
    func numberOfComponents(in pickerView: UIPickerView) -> Int
    {
        return 1
    }
    
    // The number of rows of data
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int
    {
        return oadDevices.count
    }
    
    func pickerView(_ pickerView: UIPickerView, viewForRow row: Int,
                    forComponent component: Int, reusing view: UIView?) -> UIView
    {
        let pickerLabel = UILabel()
        
        pickerLabel.text = oadDevices[row].name
        
        return pickerLabel;
    }
    
    func pickerView(_ pickerView: UIPickerView, didSelectRow row: Int,
                    inComponent component: Int)
    {
        if(oadDevices[row].name == "Custom")
        {
            addressField.isEnabled = true
        }
        else
        {
            addressField.isEnabled = false
            addressField.text = String(format: "%2X", oadDevices[row].startAddr).replacingOccurrences(of: "0x",
                                                                                                      with: "")
        }
    }
    
    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        textField.resignFirstResponder()
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {
        
    }
    
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange,
                   replacementString string: String) -> Bool
    {
        
        let separatedChar = string.components(separatedBy: hexSet)
        let filteredNums = separatedChar.joined(separator: "")
        
        if (filteredNums != string)
        {
            return false
        }
        
        if(textField == versionField && (string.count + (versionField.text?.count)!) <= 4)
        {
            return true
        }
        else if(textField == addressField && (string.count + (addressField.text?.count)!) <= 8)
        {
            return true
        }
        
        return false
    }
    
    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            let thisService = service as CBService
            
            if(service.uuid == TI_OAD_SERVICE_UUID)
            {
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        if service.uuid == TI_OAD_SERVICE_UUID
        {
            for characteristic in service.characteristics!
            {
                if(characteristic.uuid == TI_OAD_IMG_IDENTIFY_UUID)
                {
                    imgIdentChar = characteristic
                    selBTDev.setNotifyValue(true, for: characteristic)
                }
                else if(characteristic.uuid == TI_OAD_IMG_BLOCK_UUID)
                {
                    imgBlockChar = characteristic
                    selBTDev.setNotifyValue(true, for: characteristic)
                }
                else if(characteristic.uuid == TI_OAD_EXT_CTL_UUID)
                {
                    imgExtCtlChar = characteristic
                    selBTDev.setNotifyValue(true, for: characteristic)
                    timeoutTimer.invalidate()
                    connectionCon.dismiss(animated: true, completion: nil)
                }
                else
                {
                    print("OAD: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
                    continue
                }
            }
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        var values = [UInt8](repeating:0, count:characteristic.value!.count)
        characteristic.value?.copyBytes(to: &values,
                                        count: characteristic.value!.count)
        
        if(characteristic.uuid == TI_OAD_IMG_IDENTIFY_UUID)
        {
            /* Signaling to send the next block */
            metaDataRejected = true
            blockWriteSem.signal()
        }
        else if(characteristic.uuid == TI_OAD_IMG_BLOCK_UUID)
        {
            /* Signaling to send the next block */
            blockWriteSem.signal()
        }
        else if(characteristic.uuid == TI_OAD_EXT_CTL_UUID)
        {
            /* We are expecting a command for setting the block size */
            if(values[0] == FIRMWARE_OAD_BLK_SIZE_CMD)
            {
                var responseBlockSize = Int((values[2])) << 8
                responseBlockSize  = (responseBlockSize) | Int(values[1])
                firmwareDownloadView.printLogMessage(message: "Block size set to \(responseBlockSize)bytes")
                self.curBlockSize = responseBlockSize
                self.curPacketSize = responseBlockSize + 4
                
                waitForCommandSem.signal()
                
            }
        }
        else
        {
            print("OAD: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
        }
    }
    
    /* Debug log for enabling/disabling notifications */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateNotificationStateFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if(error != nil)
        {
            print("OAD ERROR: Enabling/disabling notifications failed \n ----- \(error.debugDescription)")
        }
        
        if(characteristic.isNotifying)
        {
            print("OAD: Notifications enabled for: \(characteristic.uuid.uuidString)")
        }
        else
        {
            print("OAD: Notifications disabled for: \(characteristic.uuid.uuidString)")
        }
    }
    
    /* Callback for writing a value */
    func peripheral(_ peripheral: CBPeripheral,
                    didWriteValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if(error != nil)
        {
            writeCommandFailed = true
            waitWriteSem.signal()
            blockWriteSem.signal()
        }
        else
        {
            waitWriteSem.signal()
        }
        
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        if segue.identifier == "startDownloadSegue"
        {
            let firmwareDownloadController: StatusLogViewController =
                segue.destination as! StatusLogViewController
            firmwareDownloadController.referencedController = self
            self.firmwareDownloadView = firmwareDownloadController
        }
    }
    
    /* Initiates the firmware update with the current settings */
    func initiateFirmwareUpdate()
    {
        var firmwareTarget: FirmwareTarget!
        var firmwareFormat: FirmwareFormat!
        let firmwareArray: [UInt8]
        let metadataArray: [UInt8]
        
        if(imageTypePicker.isOn)
        {
            firmwareTarget = .FIRMWARE_MSP432SAP
        }
        else
        {
            firmwareTarget = .FIRMWARE_CC2650SNP
        }
        
        if(curFirmware.filename?.contains(".txt"))!
        {
            firmwareFormat = .FIRMWARE_FORMAT_TITXT
        }
        else
        {
            firmwareFormat = .FIRMWARE_FORMAT_HEX
        }
        
        /* Initializing the semaphore we will use for synchronization between the BLE communication */
        waitWriteSem = DispatchSemaphore(value: 0)
        blockWriteSem = DispatchSemaphore(value: 0)
        waitForCommandSem = DispatchSemaphore(value: 0)
        
        /* Resetting variables */
        metaDataRejected = false
        writeCommandFailed = false
        
        self.firmwareDownloadView.printLogMessage(message: "Parsing firmware image...")
        
        /* Parsing the firmware out of the current image */
        (firmwareArray, metadataArray) = firmwareManager.prepareTXTFile(rawData: curFirmware.firmwareImageArray!,
                                                                        firmwareType: firmwareFormat,
                                                                        firmwareTarget: firmwareTarget,
                                                                        version: versionField.text!,
                                                                        address: addressField.text!)
        
        
        if(firmwareArray.count > oadDevices[devicePicker.selectedRow(inComponent: 0)].maxImageSize)
        {
            self.firmwareDownloadView.printLogMessage(message: "ERROR: Parsed firmware image exceeds maximum allowed memory for device!")
            return
        }
        
        DispatchQueue.global(qos: .userInitiated).async
            {
                self.sendFirmwareToDevice(metaImage: metadataArray, firmwareImage: firmwareArray)
                
                DispatchQueue.main.async
                    {
                        if(self.cancelInvoked)
                        {
                            
                        }
                        else if(self.writeCommandFailed == true || self.metaDataRejected == true)
                        {
                            self.firmwareDownloadView.printLogMessage(message: "Firmware update completed with errors!")
                        }
                        else
                        {
                            self.firmwareDownloadView.printLogMessage(message: "Firmware update completed successfully!")
                        }
                        
                        self.firmwareDownloadView.doneButton.setTitle("Done", for: .normal)
                }
        }
    }
    
    func sendFirmwareToDevice(metaImage: [UInt8], firmwareImage: [UInt8])
    {
        /* Setting the block size on the device */
        let blockSizeCommand = [UInt8]([FIRMWARE_OAD_BLK_SIZE_CMD,
                                        UInt8(FIRMWARE_OAD_MAX_PAYLOAD_SIZE >> 8),
                                        UInt8(FIRMWARE_OAD_MAX_PAYLOAD_SIZE & 0xFF)])
        self.selBTDev.writeValue(Data(blockSizeCommand), for: imgExtCtlChar,
                                 type: CBCharacteristicWriteType.withoutResponse)
        waitForCommandSem.wait()
        
        if(self.cancelInvoked)
        {
            return
        }
        
        /* Send the metadata to the device */
        let imageSize: UInt32 = fromByteArray([UInt8](metaImage[6..<10]), UInt32.self)
        let totalBlocks = Int(ceil((Double(imageSize*4))/Double(FIRMWARE_OAD_MAX_PAYLOAD_SIZE)))
        
        firmwareDownloadView.printLogMessage(message: "Firmware update initialized.")
        firmwareDownloadView.printLogMessage(message: "Meta Data Info:")
        firmwareDownloadView.printLogMessage(message: "    CRC32: " + String(format: "0x%X",
                                                                             fromByteArray([UInt8](metaImage[0..<4]), UInt32.self)))
        firmwareDownloadView.printLogMessage(message: "    Version: " + String(format: "0x%X",
                                                                               fromByteArray([UInt8](metaImage[4..<6]), UInt16.self)))
        firmwareDownloadView.printLogMessage(message: "    Length: \(imageSize)")
        firmwareDownloadView.printLogMessage(message: "    Address: " + String(format: "0x%X",
                                                                               fromByteArray([UInt8](metaImage[10..<14]), UInt32.self)))
        firmwareDownloadView.printLogMessage(message: "    Image type: " + String(format: "0x%X", metaImage[14]))
        firmwareDownloadView.printLogMessage(message: "    Status: " + String(format: "0x%X", metaImage[15]))
        
        self.selBTDev.writeValue(Data(metaImage), for: imgIdentChar, type: CBCharacteristicWriteType.withResponse)
        
        /* Waiting for write to clear */
        self.waitWriteSem.wait()
        
        if(self.cancelInvoked)
        {
            return
        }
        
        firmwareDownloadView.printLogMessage(message: "Performing firmware update...")
        
        /* Writing all of the data to the device */
        for ii in 0..<totalBlocks
        {
            if(self.cancelInvoked)
            {
                return
            }
            
            /* Waiting for the block request */
            blockWriteSem.wait()
            
            if(self.cancelInvoked)
            {
                return
            }
            
            /* If this variable is set it means the meta-data was rejected */
            if(metaDataRejected == true)
            {
                firmwareDownloadView.printLogMessage(message: "ERROR: Meta-data rejected! Aborting update.")
                return
            }
            else if(self.writeCommandFailed == true)
            {
                firmwareDownloadView.printLogMessage(message: "ERROR: Write command failed!")
                return
            }
            
            /* Calculating which chunk of the array to send */
            let curStartIndex = Int(ii * self.curBlockSize)
            var curEndIndex = Int()
            
            if((curStartIndex + self.curBlockSize) < firmwareImage.count)
            {
                curEndIndex = curStartIndex + self.curBlockSize
            }
            else
            {
                curEndIndex = firmwareImage.count
            }
            
            /* Constructing the packet */
            var curSendPacket = [UInt8](toByteArray(UInt32(ii)))
            curSendPacket.append(contentsOf: [UInt8](firmwareImage[curStartIndex..<curEndIndex]))
            
            while(curSendPacket.count < self.curPacketSize)
            {
                curSendPacket.append(0xFF)
            }
            
            /* Sending the packet */
            firmwareDownloadView.updateProgressBar(progress: Float(ii+1)/Float(totalBlocks))
            
            self.selBTDev.writeValue(Data(curSendPacket), for: imgBlockChar,
                                     type: CBCharacteristicWriteType.withoutResponse)
            
        }
    }
    
    func cancelUpdate()
    {
        self.cancelInvoked = true
        waitWriteSem.signal()
        blockWriteSem.signal()
        waitForCommandSem.signal()
        self.firmwareDownloadView.doneButton.setTitle("Done", for: .normal)
    }

}
