/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import Foundation
import CoreBluetooth

class OADFirmwareManager
{
    
    func prepareTXTFile(rawData: Data, firmwareType: FirmwareFormat,
                        firmwareTarget: FirmwareTarget, version: String = "FFFFFFFF",
                        address: String = "FFFF") -> (fwArray: [UInt8], metaArray: [UInt8])
    {
        var fwArray = [UInt8]()
        var metaArray = [UInt8]()
        
        /* Parsing the version/address */
        var addressArr = address.hexadecimal()!
        var versionArr = version.hexadecimal()!
        
        while(addressArr.count < 4)
        {
            addressArr.insert(0x00, at: 0)
        }
        
        while(versionArr.count < 2)
        {
            versionArr.insert(0x00, at: 0)
        }
        
        versionArr = versionArr.reversed()
        addressArr = addressArr.reversed()
        
        if(firmwareType == .FIRMWARE_FORMAT_TITXT)
        {
            var byteArrays = [[UInt8]]()
            var addresses = [UInt32]()
            var curByteArray: [UInt8]!
            var curAddress: UInt32? = nil
            
            /* Converting to a string */
            var rawString = String(data: rawData, encoding: String.Encoding.ascii)
            
            /* Removing Windows carriage returns if they are there */
            rawString = rawString?.replacingOccurrences(of: "\r\n", with: "\n")
            
            /* Splitting by new line */
            var splitString = (rawString?.components(separatedBy: "\n"))!
            
            for index in 0...((splitString.count)-1)
            {
                let curLine = splitString[index].trimmingCharacters(in: CharacterSet.whitespacesAndNewlines)
                
                if(curLine.count == 0)
                {
                    continue
                }
                
                /* Checking to see if we have an address in this line */
                if(curLine.contains("@"))
                {
                    if(curByteArray != nil)
                    {
                        byteArrays.append(curByteArray!)
                    }
                    
                    curAddress = UInt32(curLine.replacingOccurrences(of: "@", with: ""), radix: 16)!
                    addresses.append(curAddress!)
                    curByteArray = [UInt8]()
                    continue
                }
                
                /* Splitting the string and adding resulting bytes to byte array */
                let curSplitLine = curLine.components(separatedBy: " ")
                
                for curByte in curSplitLine
                {
                    if curByte == "q"
                    {
                        byteArrays.append(curByteArray)
                        continue
                    }
                    
                    curByteArray?.append(UInt8(curByte, radix: 16)!)
                }
            }
            
            /* Printing out the results */
            print("\n Firmware image parsed with the following segments: ")
            for index in 0...(addresses.count-1)
            {
                print("\n   Address: 0x\(String(format: "%X", addresses[index]))  Length: \(byteArrays[index].count)")
            }
            
            /* Calculating how long the meta-data will be */
            var metaLen: UInt32 = FIRMWARE_EXTENDED_META_LEN
            
            for index in 0...(addresses.count-1)
            {
                metaLen += FIRMWARE_EXTENDED_META_HEADER_LEN
                metaLen += UInt32(byteArrays[index].count)
            }
            
            /* Padding for MTU restrictions */
            var padSize = 0
            
            while((metaLen % 4) != 0)
            {
                metaLen += 1
                padSize += 1
            }
            
            /* Length in meta data is in words */
            metaLen = metaLen / 4
            
            /* Populating the meta data array */
            metaArray.append(contentsOf: toByteArray(UInt32(0x00000000)))          /* Placeholder for CRC */
            metaArray.append(contentsOf: versionArr)                               /* Version */
            metaArray.append(contentsOf: toByteArray(UInt32(metaLen)))             /* Length */
            metaArray.append(contentsOf: addressArr)                               /* Address */
            metaArray.append(UInt8(firmwareTarget.rawValue))                       /* Image type */
            metaArray.append(UInt8(FIRMWARE_STATUS_EXT))                           /* Status */
            
            /* Generating the OAD_ExtMeta */
            fwArray.append(contentsOf: toByteArray(UInt32(metaLen)))
            fwArray.append(contentsOf: toByteArray(UInt32(FIRMWARE_EXTENDED_FEATURE_SEGMENT)))
            fwArray.append(contentsOf: toByteArray(UInt32(FIRMWARE_SEGMENT_LIST_OFFSET)))
            fwArray.append(contentsOf: toByteArray(UInt32(0xFFFFFFFF as UInt32)))
            
            /* Generating the OAD_ExtSegmentsList */
            fwArray.append(contentsOf: toByteArray(UInt32(addresses.count)))
            
            /* Keeping track of our offsets */
            var curBinaryOffset = 20 + (UInt32(addresses.count) * FIRMWARE_EXTENDED_META_HEADER_LEN)
            
            /* Populating each segment's metadata */
            for index in 0...(addresses.count-1)
            {
                /* Offset of this segment */
                fwArray.append(contentsOf: toByteArray(UInt32(curBinaryOffset)))
                curBinaryOffset += UInt32(byteArrays[index].count)
                
                /* Address */
                fwArray.append(contentsOf: toByteArray(UInt32(addresses[index])))
                
                /* Length */
                fwArray.append(contentsOf: toByteArray(UInt32(byteArrays[index].count)))
            }
            
            /* Populating each segment's actual data */
            for index in 0...(byteArrays.count-1)
            {
                fwArray.append(contentsOf: byteArrays[index])
            }
            
            /* Adding 0xFF pads to the end of the array to make sure we are aligned */
            for _ in 0..<padSize
            {
                fwArray.append(0xFF)
            }
            
            /* TODO: Add CRC padding */
            let crc32Hash: UInt32 = crc32(0xFFFFFFFF, data: Data(fwArray))
            metaArray[0] = UInt8(crc32Hash & UInt32(0xFF))
            metaArray[1] = UInt8((crc32Hash >> 8) & UInt32(0xFF))
            metaArray[2] = UInt8((crc32Hash >> 16) & UInt32(0xFF))
            metaArray[3] = UInt8((crc32Hash >> 24) & UInt32(0xFF))
        }
        else if(firmwareType == .FIRMWARE_FORMAT_HEX)
        {
            /* Converting to a string */
            var rawString = String(data: rawData, encoding: String.Encoding.ascii)
            
            /* Keeping track of the current high */
            var curHigh: UInt16 = 0
            var lastAddr: UInt32 = 0
            var curBytes = [UInt8]()
            
            /* Removing Windows carriage returns if they are there */
            rawString = rawString?.replacingOccurrences(of: "\r\n", with: "\n")
            
            /* Splitting by new line */
            var splitString = (rawString?.components(separatedBy: "\n"))!
            var curLine = ""
            
            for index in 0...((splitString.count)-1)
            {
                /* Making sure the first line does not begin with a colon */
                if(!splitString[index].contains(":"))
                {
                    continue
                }
                else
                {
                    curLine = splitString[index].trimmingCharacters(in: CharacterSet.whitespacesAndNewlines)
                    curLine = curLine.replacingOccurrences(of: ":", with: "")
                }
                
                /* Converting into byte array */
                var curByteArray = curLine.hexadecimal()
                
                if(curByteArray == nil)
                {
                    continue
                }
                
                let curByteCount = curByteArray?[0]
                let curAddr: UInt16 = fromByteArray(Array(curByteArray![1...2]), UInt16.self).bigEndian
                let curType = curByteArray?[3]
                var curByteEnd = Int(curByteCount!) + 3
                
                if(curByteEnd >= 4)
                {
                    curBytes = Array(curByteArray![4...curByteEnd])
                }
                else
                {
                    curBytes = [UInt8]([0x00])
                }
                
                let curCRC = curByteArray?[(curByteArray?.count)!-1]

                /* Calculating the checksum */
                curByteEnd = (curByteArray?.count)! - 2
                curByteArray = Array(curByteArray![0...curByteEnd])
                var calculatedCRC: UInt8 = 0
                
                for value in curByteArray!
                {
                    let result = calculatedCRC.addingReportingOverflow(value)
                    calculatedCRC = result.0
                }
                
                calculatedCRC = calculatedCRC ^ 0xFF
                
                if(calculatedCRC == 0xFF)
                {
                    calculatedCRC = 0x00
                }
                else
                {
                    calculatedCRC = calculatedCRC + 1
                }
                
                if(calculatedCRC != curCRC!)
                {
                    return (fwArray, metaArray)
                }
                
                /* Data Type */
                if(curType! == HEX_DATA_CODE)
                {
                    let curIndexAddress = (UInt32(curHigh) << 16) | UInt32(curAddr)
                    
                    while(curIndexAddress != lastAddr)
                    {
                        fwArray.append(0xFF)
                        lastAddr = lastAddr + 1
                    }
                    
                    fwArray.append(contentsOf: curBytes)
                    lastAddr = curIndexAddress + UInt32(curByteCount!)
                    
                }
                else if(curType == HEX_EOF_CODE)
                {
                    break
                }
                else if(curType == HEX_EXT_LINN_ADDR_CODE)
                {
                    curHigh = fromByteArray(curBytes, UInt16.self).bigEndian
                }
            }
            
            while((fwArray.count % 4) != 0)
            {
                fwArray.append(0xFF)
            }
            
            print("FirmwareParser: Firmware parsed with size \(fwArray.count) bytes")
            
            /* Populating the meta data array */
            let crc32Hash: UInt32 = crc32(0xFFFFFFFF, data: Data(fwArray))
            metaArray.append(contentsOf: toByteArray(UInt32(crc32Hash)))               /*  CRC */
            metaArray.append(contentsOf: versionArr)                                   /* Version */
            metaArray.append(contentsOf: toByteArray(UInt32(fwArray.count/4)))         /* Length */
            metaArray.append(contentsOf: addressArr)                                   /* Address */
            metaArray.append(UInt8(firmwareTarget.rawValue))                           /* Image type */
            metaArray.append(UInt8(FIRMWARE_STATUS_EXT))                               /* Status */
            
        }
        
        return (fwArray, metaArray)
    }

}
